@@ -50,18 +50,23 @@
 #include <sys/file.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+#include <netinet/in.h>
 #include <rpc/rpc.h>
 #include <rpc/rpc_com.h>
 #ifdef PORTMAP
 #include <netinet/in.h>
 #endif
 #include <arpa/inet.h>
+#ifdef SYSTEMD
+#include <systemd/sd-daemon.h>
+#endif
 #include <fcntl.h>
 #include <netdb.h>
 #include <stdio.h>
 #include <netconfig.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <grp.h>
 #include <syslog.h>
 #include <err.h>
 #include <pwd.h>
@@ -73,6 +78,7 @@
 static inline void __nss_configure_lookup(const char *db, const char *s) {}
 #endif
 #include "rpcbind.h"
+#include "xlog.h"
 
 /*#define RPCBIND_DEBUG*/
 
@@ -81,6 +87,7 @@ static inline void __nss_configure_lookup(const char *db, const char *s) {}
 int debugging = 0;	/* Tell me what's going on */
 int doabort = 0;	/* When debugging, do an abort on errors */
 int dofork = 1;		/* fork? */
+int createdsocket = 0;  /* Did I create the socket or systemd did it for me? */
 
 rpcblist_ptr list_rbl;	/* A list of version 3/4 rpcbind services */
 
@@ -90,6 +97,12 @@ char *rpcbinduser = RPCBIND_USER;
 char *rpcbinduser = "";
 #endif
 
+#ifdef NSS_MODULES
+char *nss_modules = NSS_MODULES;
+#else
+char *nss_modules = "files";
+#endif
+
 /* who to suid to if -s is given */
 #ifdef __CYGWIN__
 #define RUN_AS  "SYSTEM"
@@ -109,6 +122,9 @@ int nhosts = 0;
 int on = 1;
 int rpcbindlockfd;
 
+#ifdef LIB_SET_DEBUG
+void  libtirpc_set_debug(char *name, int level, int use_stderr);
+#endif
 #ifdef WARMSTART
 /* Local Variable */
 static int warmstart = 0;	/* Grab an old copy of registrations. */
@@ -124,13 +140,13 @@ char *tcp_uaddr;	/* Universal TCP address */
 static char servname[] = "rpcbind";
 static char superuser[] = "superuser";
 
-int main __P((int, char *[]));
+int main(int, char *[]);
 
-static int init_transport __P((struct netconfig *));
-static void rbllist_add __P((rpcprog_t, rpcvers_t, struct netconfig *,
-			     struct netbuf *));
-static void terminate __P((int));
-static void parseargs __P((int, char *[]));
+static int init_transport(struct netconfig *);
+static void rbllist_add(rpcprog_t, rpcvers_t, struct netconfig *,
+			     struct netbuf *);
+static void terminate(int);
+static void parseargs(int, char *[]);
 
 int
 main(int argc, char *argv[])
@@ -158,7 +174,7 @@ main(int argc, char *argv[])
 			rl.rlim_cur = 128;
 		setrlimit(RLIMIT_NOFILE, &rl);
 	}
-	openlog("rpcbind", LOG_CONS, LOG_DAEMON);
+
 #ifndef __CYGWIN__
 	if (geteuid()) { /* This command allowed only to root */
 		fprintf(stderr, "Sorry. You are not superuser\n");
@@ -170,7 +186,7 @@ main(int argc, char *argv[])
 	 * Make sure we use the local service file 
 	 * for service lookkups
 	 */
-	__nss_configure_lookup("services", "files");
+	__nss_configure_lookup("services", nss_modules);
 
 	nc_handle = setnetconfig(); 	/* open netconfig file */
 	if (nc_handle == NULL) {
@@ -185,7 +201,20 @@ main(int argc, char *argv[])
 		syslog(LOG_ERR, "%s: can't find local transport\n", argv[0]);
 		exit(1);
 	}
-	
+	xlog_open("rpcbind");
+	if (dofork) {
+		xlog_syslog(TRUE);
+		xlog_stderr(FALSE);
+	} else {
+		xlog_syslog(FALSE);
+		xlog_stderr(TRUE);
+	}
+	if (debugging) {
+		xlog_config(D_ALL, 1);
+#ifdef LIB_SET_DEBUG
+		libtirpc_set_debug("rpcbind", debugging, (dofork == 0));
+#endif
+	}
 	rpc_control(RPC_SVC_CONNMAXREC_SET, &maxrec);
 
 	init_transport(nconf);
@@ -214,18 +243,16 @@ main(int argc, char *argv[])
 	(void) signal(SIGUSR1, SIG_IGN);
 	(void) signal(SIGUSR2, SIG_IGN);
 
-	if (debugging) {
 #ifdef RPCBIND_DEBUG 
-		printf("rpcbind debugging enabled.");
-		if (doabort) {
-			printf("  Will abort on errors!\n");
-		} else {
-			printf("\n");
-		}
+	if (debugging) {
+		xlog(LOG_DEBUG, "debugging enabled.");
+		if (doabort)
+			xlog(LOG_DEBUG, "Will abort on errors!\n");
+	}
 #endif
-	} else if (dofork) {
+	if (dofork) {
 		if (daemon(0, 0))
-        		err(1, "fork failed");
+        	err(1, "fork failed");
 	}
 
 	if (runasdaemon || *rpcbinduser) {
@@ -236,12 +263,17 @@ main(int argc, char *argv[])
 		 * Make sure we use the local password file
 		 * for these lookups.
 		 */
-		__nss_configure_lookup("passwd", "files");
+		__nss_configure_lookup("passwd", nss_modules);
 
 		if((p = getpwnam(id)) == NULL) {
 			syslog(LOG_ERR, "cannot get uid of '%s': %m", id);
 			exit(1);
 		}
+#ifdef WARMSTART
+		if (warmstart) {
+			mkdir_warmstart(p->pw_uid);
+		}
+#endif
                 if (setgid(p->pw_gid) == -1) {
                         syslog(LOG_ERR, "setgid to '%s' (%d) failed: %m", id, p->pw_gid);
                         exit(1);
@@ -264,6 +296,13 @@ main(int argc, char *argv[])
 
 	network_init();
 
+#ifdef SYSTEMD
+	/* Try to notify system of successful startup, regardless of whether we
+	 * used systemd socket activation or not. When started from the command
+	 * line, this should not hurt either.
+	 */
+	sd_notify(0, "READY=1");
+#endif
 	my_svc_run();
 	syslog(LOG_ERR, "svc_run returned unexpectedly");
 	rpcbind_abort();
@@ -272,6 +311,31 @@ main(int argc, char *argv[])
 	return 0;
 }
 
+/*
+ * Normally systemd will open sockets in dual ipv4/ipv6 mode.
+ * That won't work with netconfig and we'll only match
+ * the ipv6 socket. Convert it to IPV6_V6ONLY and issue
+ * a warning for the user to fix their systemd config.
+ */
+static int
+handle_ipv6_socket(int fd)
+{
+	int opt;
+	socklen_t len = sizeof(opt);
+
+	if (getsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &opt, &len)) {
+		syslog(LOG_ERR, "failed to get ipv6 socket opts: %m");
+		return -1;
+	}
+
+	if (opt) /* socket is already in V6ONLY mode */
+		return 0;
+
+	syslog(LOG_ERR, "systemd has passed an IPv4/IPv6 dual-mode socket.");
+	syslog(LOG_ERR, "Please fix your systemd config by specifying IPv4 and IPv6 sockets separately and using BindIPv6Only=ipv6-only.");
+	return -1;
+}
+
 /*
  * Adds the entry into the rpcbind database.
  * If PORTMAP, then for UDP and TCP, it adds the entries for version 2 also
@@ -295,6 +359,7 @@ init_transport(struct netconfig *nconf)
 	u_int32_t host_addr[4];  /* IPv4 or IPv6 */
 	struct sockaddr_un sun;
 	mode_t oldmask;
+	int n;
         res = NULL;
 
 	if ((nconf->nc_semantics != NC_TPI_CLTS) &&
@@ -306,37 +371,93 @@ init_transport(struct netconfig *nconf)
 		int i;
 		char **s;
 
-		(void) fprintf(stderr, "%s: %ld lookup routines :\n",
+		(void) xlog(LOG_DEBUG, "%s: %ld lookup routines :\n",
 			nconf->nc_netid, nconf->nc_nlookups);
 		for (i = 0, s = nconf->nc_lookups; i < nconf->nc_nlookups;
 		     i++, s++)
-			fprintf(stderr, "[%d] - %s\n", i, *s);
+			xlog(LOG_DEBUG, "[%d] - %s\n", i, *s);
 	}
 #endif
-
-	/*
-	 * XXX - using RPC library internal functions. For NC_TPI_CLTS
-	 * we call this later, for each socket we like to bind.
-	 */
-	if (nconf->nc_semantics != NC_TPI_CLTS) {
-		if ((fd = __rpc_nconf2fd(nconf)) < 0) {
-			syslog(LOG_ERR, "cannot create socket for %s",
-			    nconf->nc_netid);
-			return (1);
-		}
-	}
-
 	if (!__rpc_nconf2sockinfo(nconf, &si)) {
 		syslog(LOG_ERR, "cannot get information for %s",
 		    nconf->nc_netid);
 		return (1);
 	}
 
+#ifdef SYSTEMD
+	n = sd_listen_fds(0);
+	if (n < 0) {
+		syslog(LOG_ERR, "failed to acquire systemd sockets: %s", strerror(-n));
+		return 1;
+	}
+
+	/* Try to find if one of the systemd sockets we were given match
+	 * our netconfig structure. */
+
+	for (fd = SD_LISTEN_FDS_START; fd < SD_LISTEN_FDS_START + n; fd++) {
+		struct __rpc_sockinfo si_other;
+		union {
+			struct sockaddr sa;
+			struct sockaddr_un un;
+			struct sockaddr_in in4;
+			struct sockaddr_in6 in6;
+			struct sockaddr_storage storage;
+		} sa;
+		socklen_t addrlen = sizeof(sa);
+
+		if (!__rpc_fd2sockinfo(fd, &si_other)) {
+			syslog(LOG_ERR, "cannot get information for fd %i", fd);
+			return 1;
+		}
+
+		if (si.si_af != si_other.si_af ||
+                    si.si_socktype != si_other.si_socktype ||
+                    si.si_proto != si_other.si_proto)
+			continue;
+
+		if (getsockname(fd, &sa.sa, &addrlen) < 0) {
+			syslog(LOG_ERR, "failed to query socket name: %s",
+                               strerror(errno));
+			goto error;
+		}
+
+		if (sa.sa.sa_family == AF_INET6 && handle_ipv6_socket(fd))
+		        goto error;
+
+		/* Copy the address */
+		taddr.addr.maxlen = taddr.addr.len = addrlen;
+		taddr.addr.buf = malloc(addrlen);
+		if (taddr.addr.buf == NULL) {
+			syslog(LOG_ERR,
+                               "cannot allocate memory for %s address",
+                               nconf->nc_netid);
+			goto error;
+		}
+		memcpy(taddr.addr.buf, &sa, addrlen);
+
+		my_xprt = (SVCXPRT *)svc_tli_create(fd, nconf, &taddr,
+                          RPC_MAXDATASIZE, RPC_MAXDATASIZE);
+		if (my_xprt == (SVCXPRT *)NULL) {
+			syslog(LOG_ERR, "%s: could not create service",
+                               nconf->nc_netid);
+			goto error;
+		}
+	}
+
+	/*
+	 * If none of the systemd sockets matched, we set up the socket in
+	 * the normal way:
+	 */
+#endif
+	if (my_xprt != NULL)
+		goto got_socket;
+
 	if ((strcmp(nconf->nc_netid, "local") == 0) ||
 	    (strcmp(nconf->nc_netid, "unix") == 0)) {
 		memset(&sun, 0, sizeof sun);
 		sun.sun_family = AF_LOCAL;
 		unlink(_PATH_RPCBINDSOCK);
+		createdsocket = 1; /* We are now in the process of creating the unix socket */
 		strcpy(sun.sun_path, _PATH_RPCBINDSOCK);
 		addrlen = SUN_LEN(&sun);
 		sa = (struct sockaddr *)&sun;
@@ -382,11 +503,13 @@ init_transport(struct netconfig *nconf)
 				    nconf->nc_netid);
 				return (1);
 			}
+
+			hints.ai_flags &= ~AI_NUMERICHOST;
 			switch (hints.ai_family) {
 			case AF_INET:
 				if (inet_pton(AF_INET, hosts[nhostsbak],
 				    host_addr) == 1) {
-					hints.ai_flags &= AI_NUMERICHOST;
+					hints.ai_flags |= AI_NUMERICHOST;
 				} else {
 					/*
 					 * Skip if we have an AF_INET6 adress.
@@ -399,7 +522,7 @@ init_transport(struct netconfig *nconf)
 			case AF_INET6:
 				if (inet_pton(AF_INET6, hosts[nhostsbak],
 				    host_addr) == 1) {
-					hints.ai_flags &= AI_NUMERICHOST;
+					hints.ai_flags |= AI_NUMERICHOST;
 				} else {
 					/*
 					 * Skip if we have an AF_INET adress.
@@ -476,7 +599,7 @@ init_transport(struct netconfig *nconf)
 				}
 				nb.len = nb.maxlen = sa_size;
 				uaddr = taddr2uaddr(nconf, &nb);
-				(void) fprintf(stderr,
+				(void) xlog(LOG_DEBUG,
 				    "rpcbind : my address is %s\n", uaddr);
 				(void) free(uaddr);
 			}
@@ -492,6 +615,12 @@ init_transport(struct netconfig *nconf)
 		if (!checkbind)
 			return 1;
 	} else {	/* NC_TPI_COTS */
+		if ((fd = __rpc_nconf2fd(nconf)) < 0) {
+			syslog(LOG_ERR, "cannot create socket for %s",
+			    nconf->nc_netid);
+			return (1);
+		}
+
 		if ((strcmp(nconf->nc_netid, "local") != 0) &&
 		    (strcmp(nconf->nc_netid, "unix") != 0)) {
 			if ((aicode = getaddrinfo(NULL, servname, &hints, &res))!= 0) {
@@ -553,7 +682,7 @@ init_transport(struct netconfig *nconf)
 			}
 			nb.len = nb.maxlen = sa_size2;
 			uaddr = taddr2uaddr(nconf, &nb);
-			(void) fprintf(stderr, "rpcbind : my address is %s\n",
+			(void) xlog(LOG_DEBUG, "rpcbind : my address is %s\n",
 			    uaddr);
 			(void) free(uaddr);
 		}
@@ -568,6 +697,7 @@ init_transport(struct netconfig *nconf)
 			goto error;
 		}
 	}
+got_socket:
 
 #ifdef PORTMAP
 	/*
@@ -659,36 +789,40 @@ init_transport(struct netconfig *nconf)
 #ifdef RPCBIND_DEBUG
 	if (debugging) {
 		if (status < 0) {
-			fprintf(stderr, "Error in finding bind status for %s\n",
+			xlog(LOG_DEBUG, "Error in finding bind status for %s\n",
 				nconf->nc_netid);
 		} else if (status == 0) {
-			fprintf(stderr, "check binding for %s\n",
+			xlog(LOG_DEBUG, "check binding for %s\n",
 				nconf->nc_netid);
 		} else if (status > 0) {
-			fprintf(stderr, "No check binding for %s\n",
+			xlog(LOG_DEBUG, "No check binding for %s\n",
 				nconf->nc_netid);
 		}
 	}
 #endif
+
+
+#ifdef RMTCALLS
 	/*
 	 * rmtcall only supported on CLTS transports for now.
 	 */
 	if (nconf->nc_semantics == NC_TPI_CLTS) {
 		status = create_rmtcall_fd(nconf);
-
 #ifdef RPCBIND_DEBUG
 		if (debugging) {
 			if (status < 0) {
-				fprintf(stderr,
+				xlog(LOG_DEBUG,
 				    "Could not create rmtcall fd for %s\n",
 					nconf->nc_netid);
 			} else {
-				fprintf(stderr, "rmtcall fd for %s is %d\n",
+				xlog(LOG_DEBUG, "rmtcall fd for %s is %d\n",
 					nconf->nc_netid, status);
 			}
 		}
 #endif
 	}
+#endif
+
 	return (0);
 error:
 	close(fd);
@@ -708,8 +842,8 @@ rbllist_add(rpcprog_t prog, rpcvers_t vers, struct netconfig *nconf,
 	}
 #ifdef RPCBIND_DEBUG	
 	if (debugging){
-	  fprintf(stderr,"FUNCTION rbllist_add");
-	  fprintf(stderr,"Add the prog %lu vers %lu to the rpcbind list\n",
+	  xlog(LOG_DEBUG,"FUNCTION rbllist_add");
+	  xlog(LOG_DEBUG,"Add the prog %lu vers %lu to the rpcbind list",
                   (ulong)prog, (ulong)vers);
 	}
 #endif	
@@ -729,14 +863,13 @@ static void
 terminate(int dummy /*__unused*/)
 {
 	close(rpcbindlockfd);
-	unlink(_PATH_RPCBINDSOCK);
+	if(createdsocket)
+		unlink(_PATH_RPCBINDSOCK);
 	unlink(RPCBINDDLOCK);
 #ifdef WARMSTART
-	syslog(LOG_ERR,
-		"rpcbind terminating on signal. Restart with \"rpcbind -w\"");
 	write_warmstart();	/* Dump yourself */
 #endif
-	exit(2);
+	exit(0); /* exit gracefully */
 }
 
 void
@@ -761,7 +894,7 @@ parseargs(int argc, char *argv[])
 			break;		/* errors; for rpcbind developers */
 					/* only! */
 		case 'd':
-			debugging = 1;
+			debugging++;
 			break;
 		case 'h':
 			++nhosts;
